/*
Network motif analysis tools
Copyright (C) 2006 Johannes Knabe
j.f.knabe@herts.ac.uk

Free for you to use in any way, but acknowledgement / reference has to be given.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
/*
network class methods:

public network();
public network(int[][] netwrk);
public node addNode(String name, int capacity);
public node addNode(String name);
public node addNode();
public void addNodes(int howmany);
public void swapNodes(int i, int j);
public void link(int fromNum, int toNum, boolean inhibitory);
public void link(int fromNum, int toNum);
public network partition(int[] how);
public double countWithinSubclassEdges(int subclass);
public double subclassDegreeSum(int subclass);
public double getModularity();
public void printConnections();
public int[][] getConnectivityMatrixArray(boolean considerInhibitory);
public int[][] getConnectivityMatrixArray();
public int[][] getSubclassConnectivityMatrixArray(int subclass, boolean considerInhibitory);
public int[][] getSubclassConnectivityMatrixArray(int subclass);
public void printConnectivityMatrix(boolean inhibitory);
public void printConnectivityMatrix();

static network methods:

public static Iterable<network> getAllPartitions(final network netwrk);
public static int[][] randomizeConnectivityMatrix(int[][] newrk);
public static network getNetworkFromConnMat(int[][] netwrk);
public static int countEdges(final int[][] nework);
public static double getDensity(final int[][] nework);
public static double[] getMaximumModularity(final int[][] nework);
public static String getStringRep(final int[][] nework);
public static String getCanonicalLabel(final int[][] nework);
public static int getIntRep(final int[][] nework);
public static int getCanonicalInt(final int[][] nework);
public static int[][] canonicalIntToConnMat(final int canon, final int squareSize);
public static int[][] canonicalIntToConnMat(final int canon);
public static Iterable<int[][]> getAllPermutations(final int[][] netwrk);
public static int[][] getSubMatrix(int[][] newrk, int[] in);
public static Iterable<int[][]> getAllNbyNsubMatrices(final int n, final int[][] netwrk);
public static void printConnectivityMatrix(int[][] newrk);
public static void printConnectivityMatrix(int[][] newrk, boolean inhibitory);
*/
import java.util.*;

public class network{

public static java.util.Random rand=new java.util.Random(230581);
public static int[] lastIn;

public Vector<node> nodes;

public int numberOfSubclasses=0;   //no need to give any subclasses, is currently used only when partitions are present, see "partition" function
public int[] currentPartition=null;//no need to give this, see above "numberOfSubclasses"
public network(int size){
  this.nodes=new Vector<node>(size);
}

public network(){
  this(10);
}

//construct network from connecitvity matrix
public network(int[][] netwrk){
  this(netwrk.length);
  this.addNodes(netwrk.length);
  for (int i=0;i<netwrk.length;i++)
   for (int j=0;j<netwrk[i].length;j++)
    if(netwrk[i][j]!=0){
     if(netwrk[i][j]<0)
      link(i,j,true);//is inhibitory
     else
      link(i,j);
     }
}

public node addNode(String name, int capacity){
  node newnode=new node(name, capacity);
  this.nodes.add(newnode);
  return newnode;
}

public node addNode(String name){
  return addNode(name, nodes.capacity());
}

public node addNode(){
  return addNode(""+nodes.size());
}

public void addNodes(int howmany){
  for(int i=0; i<howmany;i++)
    addNode();
}

public void swapNodes(int i, int j){
 if(i!=j){
  node t1=this.nodes.elementAt(i);
  node t2=this.nodes.elementAt(j);
  this.nodes.setElementAt(t1,j);
  this.nodes.setElementAt(t2,i);
 }
}

public void link(int fromNum, int toNum, boolean inhibitory){
  node from=nodes.elementAt(fromNum);
  from.linkto(nodes.elementAt(toNum),inhibitory);
}

public void link(int fromNum, int toNum){
  link(fromNum, toNum, false);
}

//partitions the network into subnetworks, e.g. to make a three node net into 
//a one two- and a one-node net you would call partition(new int[]{2,1}); (or partition(new int[]{1,2});, nodes are taken and grouped as they come from the nodes iterator!)
//note that this does not create a new network, so it returns only the modified current network
public network partition(int[] how){
int sum=0;
for(int i=0;i<how.length;i++)
  sum+=how[i];
if(sum<nodes.size()){
  System.err.println("ERROR: partition has to be called with a partitioning that considers all nodes!");
  System.exit(1);
  }
numberOfSubclasses=how.length;
currentPartition=new int[how.length];
System.arraycopy(how,0,currentPartition,0,how.length);
int t=0;
for (node current: nodes){
 current.subclass=t;
 how[t]--;
 if(how[t]==0)
   t++;
 }
return this;
}

//gets you one after the other all possible partitions of the network, 
//e.g. for a three node network: partition({3}),partition({2,1}),partition({1,2}),partition({1,1,1})
//this includes order, partition({2,1}) and partition({1,2}) are not considered the same!
public static Iterable<network> getAllPartitions(final network netwrk){
  return new Iterable<network>() {
    public Iterator<network> iterator(){
      return new Iterator<network>() {
        public final network newrk=netwrk;
        private partitionIterator t=new partitionIterator(newrk.nodes.size());
        public boolean gotMore=true;
        public int[] currentPartition;
        public boolean hasNext(){return gotMore;}
        public network next(){
          currentPartition=t.next();
          gotMore=false;
          for(int i=0;i<currentPartition.length;i++)
            if(currentPartition[i]>1)
              gotMore=true;
          return newrk.partition(currentPartition);
          }
        public void remove(){}
        };
      }
    };
}

//gives the sum of all edges that lie within the given subclass (module)
public double countWithinSubclassEdges(int subclass){
int sum=0;
for (node current: nodes){
  if(current.subclass==subclass)
    sum+=current.countWithinSubclassEdges();
  }
return sum;
}

//gives the sum of out-degrees of all nodes in the given subclass (module),
//regardless of whether they connect within or without that subclass
public double subclassDegreeSum(int subclass){
int sum=0;
for (node current: nodes){
  if(current.subclass==subclass)
    sum+=current.outgoing.size();
  }
return sum;
}

//For a given partition of the network, this will calculate the modularity of the network
//according to formulas given in http://www.nature.com/nature/journal/v433/n7028/pdf/nature03288.pdf
public double getModularity(){
double sum=0;
double edges=network.countEdges(this.getConnectivityMatrixArray());
for(int s=0;s<numberOfSubclasses;s++){
  double temp=this.subclassDegreeSum(s)/(edges);//factor of 2 in the paper does not apply here!
  sum+=(this.countWithinSubclassEdges(s)/edges)-(temp*temp);
  }
return sum;
}

public void printConnections(){
  System.out.println("Printing list of all edges:");
  for (node current: nodes)
    current.printConnections();
}

//connectivity matrix of the network, e.g. for
//two nodes 0 and 1 where both are self-connected and
//0 inputs to 1 but not the other way round:
//  | 0 | 1
//----------
// 0| 1 | 1
// ---------
// 1| 0 | 1
//(output is connections only, without node names)
public int[][] getConnectivityMatrixArray(boolean considerInhibitory){
  int[][] ret =new int[nodes.size()][nodes.size()];
  int i=0; int j=0;
  for (node current: nodes){
   for (node inputToThis: nodes){
    ret[i][j]=current.isOutgoing(inputToThis,considerInhibitory);
    j++;
    }
   i++; j=0;
   }
  return ret;
}

public int[][] getConnectivityMatrixArray(){
  return getConnectivityMatrixArray(false);
}

public int[][] getSubclassConnectivityMatrixArray(int subclass, boolean considerInhibitory){
  int[][] ret =new int[currentPartition[subclass]][currentPartition[subclass]];
  int i=0; int j=0;
  for (node current: nodes)
   if(current.subclass==subclass){
    for (node inputToThis: nodes)
     if(inputToThis.subclass==subclass){
      ret[i][j]=current.isOutgoing(inputToThis,considerInhibitory);
      j++;
      }
     i++; j=0;
   }
  return ret;
}

public int[][] getSubclassConnectivityMatrixArray(int subclass){
return getSubclassConnectivityMatrixArray(subclass,false);
}

public void printConnectivityMatrix(boolean inhibitory){
  printConnectivityMatrix(getConnectivityMatrixArray(inhibitory),inhibitory);
}

public void printConnectivityMatrix(){
  printConnectivityMatrix(false);
}
///////////////////////////////////////
//begin static network functions
///////////////////////////////////////

//this randomization conserves the numbers of 0 and 1 in every row and column
//(for example a dead node will stay death!)
//described e.g. by Shai S. Shenn-Orr, Ron Milo, Shmoolik Mangan & Uri Alon 2002
public static int[][] randomizeConnectivityMatrix(int[][] newrk){
int n=0;
int i=0;
int x1,y1,x2,y2;
int matrixSize=(newrk.length*newrk[0].length);
int cubicMatrixSize;
if(matrixSize>=400)
  cubicMatrixSize=64000000;
else
  cubicMatrixSize=matrixSize*matrixSize*matrixSize;
while((n<=matrixSize*3)&&(i<=cubicMatrixSize)){
  do{i++;                                     //find first rand.nextIntom connection
    x1=rand.nextInt(newrk.length);
    y1=rand.nextInt(newrk.length);
    }  while(newrk[x1][y1]==0);
  do{i++;                                     //find second rand.nextIntom connection
    x2=rand.nextInt(newrk.length);
    y2=rand.nextInt(newrk.length);
    }  while(newrk[x2][y2]==0);
  i++;
  if((newrk[x1][y2]==0)&&(newrk[x2][y1]==0)){ //swap connections if they do not exist yet
      newrk[x1][y2]=newrk[x1][y1];
      newrk[x2][y1]=newrk[x2][y2];
      newrk[x1][y1]=0;
      newrk[x2][y2]=0;
      n++;
    }
  }
if(i>=cubicMatrixSize){
  System.out.println("ERRROR! Could not really randomize ");
  printConnectivityMatrix(newrk);
  System.out.println("Given up with an i of "+i+"!!!");
  //System.exit(1);
  }
return newrk;
}

public static network getNetworkFromConnMat(int[][] netwrk){
return new network(netwrk);
}

public static int countEdges(final int[][] nework){
int sum=0;
for (int i=0;i<nework.length;i++)
 for (int j=0;j<nework[i].length;j++)
  if(nework[i][j]!=0)
   sum+=1;
return sum;
}

//returns the graph's density D = |E| / |V|^2 (edges divided by nodes times nodes),
//so we get a value between 0 and 1
public static double getDensity(final int[][] nework){
return ((double)countEdges(nework))/(nework.length*nework.length*1.0);
}

//maximizes a modularity measure as defined in:
//http://www.nature.com/nature/journal/v433/n7028/pdf/nature03288.pdf
//The return array contains, 1) that maximum modularity value and 
//                           2) the partition array used for this maximum
//ALL combinations are tried, so do not do this for big matrizes!!!
public static double[] getMaximumModularity(final int[][] nework){
network t=new network(nework);
int[] usedPartition=null;
double maxMod=-1.0;
for(network current: getAllPartitions(t)){
  double tempMod=current.getModularity();
  if(tempMod>maxMod){
    maxMod=tempMod;
    usedPartition=new int[current.currentPartition.length];
    System.arraycopy(current.currentPartition,0,usedPartition,0,current.currentPartition.length);
    }
  }
double res[]=new double[usedPartition.length+1];
res[0]=maxMod;
for(int i=0;i<usedPartition.length;i++)
  res[i+1]=usedPartition[i];
return res;
}

public static int factorial( int n )
{
  if( n <= 1 )
   return 1;
  else
   return n * factorial( n - 1 );
}

//concatenation of all rows of the connectivity matrix
public static String getStringRep(final int[][] nework){
String res="";
for (int i=0;i<nework.length;i++)
 for (int j=0;j<nework[i].length;j++)
  res+=nework[i][j];
return res;
}

public static String getCanonicalLabel(final int[][] nework){
String smallest="999999999999999999999999999999999999999999999999999999999999999999999999";
for(int[][] newrk: network.getAllPermutations(nework)){
  String t=getStringRep(newrk);
  if(t.compareTo(smallest)<0)
    smallest=t;
  }
return smallest;
}


public static int getIntRep(final int[][] nework){
if(nework.length>5){
  System.err.println("Only for matrizes up to 5*5 size because of 32 bit limit");
  System.exit(1);
  }
int res=0;
for (int i=0;i<nework.length;i++)
 for (int j=0;j<nework[i].length;j++)
  res=res | (Math.abs(nework[i][j]) << i*nework.length+j);
return res;
}

public static int getCanonicalInt(final int[][] nework){
int smallest=Integer.MAX_VALUE;
for(int[][] newrk: network.getAllPermutations(nework)){
  int t=getIntRep(newrk);
  if(t<smallest)
    smallest=t;
  }
return smallest;
}

//This will return the connectivity matrix represented by this integer
//as small matrizes might start with a lot of 0s the squareSize should be given
//(e.g. squareSize=9 for a 3 node net)
public static int[][] canonicalIntToConnMat(final int canon, final int squareSize){
String bits=Integer.toBinaryString(canon);
while(bits.length()<squareSize)
  bits="0"+bits;
int netSize=(int) Math.sqrt(squareSize);
int[][] res=new int[netSize][netSize];
int c=0;
for(int i=0;i<netSize;i++)
  for(int j=0;j<netSize;j++){
    res[i][j]=Integer.parseInt(bits.charAt(c)+"");
    c++;
  }
return res;
}

public static int[][] canonicalIntToConnMat(final int canon){
System.err.println("WARNING: You will most probably losse precision here! Please give expected size to allow filling with 0s!");
String bits=Integer.toBinaryString(canon);
return canonicalIntToConnMat(canon,bits.length());
}

//gives the conn. matrizes of all permutations of node orderings of the network, e.g. [1,2,3],[2,3,1],...
public static Iterable<int[][]> getAllPermutations(final int[][] netwrk){
  return new Iterable<int[][]>() {
    public Iterator<int[][]> iterator(){
      return new Iterator<int[][]>() {
        public final int[][] newrk=netwrk;
        public final int lim=factorial(netwrk.length);
        public int i=0;
        public boolean hasNext(){return i<lim;}
        public void permutation(network s, int k, int n) {
          int factorial= 1;
          for(int j= 2;j<n;j++){
            factorial= factorial* (j-1);
            s.swapNodes( j-1- ((k / factorial) % j), j-1);
          }
        }
        public int[][] next(){
          network t=new network(newrk);
          permutation(t,i,newrk.length+1);
          i++;
          return t.getConnectivityMatrixArray();}
        public void remove(){}
        };
      }
    };
}


public static int[][] getSubMatrix(int[][] newrk, int[] in){
 lastIn=new int[in.length];
 System.arraycopy(in,0,lastIn,0,in.length);
 int[][] ret=new int[in.length][in.length];
 try{
  for (int i=0;i<in.length;i++)
   for (int j=0;j<in.length;j++)
    ret[i][j]=newrk[in[i]][in[j]];
 }catch(Exception e){System.out.print("Invalid columns/rows for submatrix; you provided:"); for(int i=0;i<in.length;i++) System.out.print(" "+in[i]); System.out.println();  System.exit(1);}
 //System.out.print("Submartix:"); for(int i=0;i<in.length;i++) System.out.print(" "+in[i]); System.out.println();
 return ret;
}

public static Iterable<int[][]> getAllNbyNsubMatrices(final int n, final int[][] netwrk){
  return new Iterable<int[][]>() {
    public Iterator<int[][]> iterator(){
      return new Iterator<int[][]>() {
        public final int num=n;
        public boolean finalOne=false;
        public int[] in=new int[n]; //rows / colums that are in the current submatrix
        {
        for(int i=0;i<n;i++)
           in[i]=i;
        }
        public final int[][] newrk=netwrk;
        public boolean hasNext(){if(finalOne) return false; if(in[0]>=newrk.length-n) finalOne=true; return true;}
        public int[][] next(){
          int[][] ret=getSubMatrix(newrk,in);
          for(int i=n-1;i>=0;i--)
           if(in[i]<newrk.length-(n-i))
            {in[i]++;
             for(int j=n-1;j>i;j--)
              in[j]=in[j-1]+1;
             break;}
          return ret;}
        public void remove(){}
        };
      }
    };
}

public static void printConnectivityMatrix(int[][] newrk){
printConnectivityMatrix(newrk,false);
}

public static void printConnectivityMatrix(int[][] newrk, boolean inhibitory){
  System.out.println("Printing Connectivity Matrix:");
  for (int i=0;i<newrk.length;i++){
   for (int j=0;j<newrk[i].length;j++){
     if(inhibitory){
      if(newrk[i][j]<0)
       System.out.print(newrk[i][j]);
      else
       System.out.print(" "+newrk[i][j]);
     }
     else
      System.out.print(" "+Math.abs(newrk[i][j]));
     }
    System.out.println();
   }
}
///////////////////////////////////////
//end static network functions
///////////////////////////////////////

}